from pathlib import Path

from playwright.sync_api import Page, expect

from pynanto import Webserver
from tests import for_all_webservers, new_config


@for_all_webservers()
def test_quickstart_bootstrap(page: Page, webserver: Webserver):
    config = new_config(webserver)
    config.bootstrap.add_python(
        """print('running bootstrap custom code')\n"""
        """from js import document\n"""
        """document.body.innerHTML='<input id="tag1" value="Hello world 1!">'\n"""
        """print(f'document.body.innerHTML=[{document.body.innerHTML}]')\n"""
    )
    page.goto(config.webserver.localhost_url())
    expect(page.locator('id=tag1')).to_have_value('Hello world 1!')


@for_all_webservers()
def test_bootstrap_bundle(page: Page, webserver: Webserver):
    config = new_config(webserver)
    config.bundles.add_file_content(
        'script1.py',
        """from js import document\n"""
        """document.body.innerHTML='<input id="tag1" value="Hello world 2!">'\n"""
    )
    config.bootstrap.add_python('import script1')

    page.goto(config.webserver.localhost_url())
    expect(page.locator('id=tag1')).to_have_value('Hello world 2!')


@for_all_webservers()
def test_bootstrap_bundle_from_filesystem(page: Page, webserver: Webserver):
    config = new_config(webserver)
    config.bundles.add_flat_folder(Path(__file__).parent / 'data' / 'for_remote')
    config.bootstrap.add_python('import main')

    page.goto(config.webserver.localhost_url())
    expect(page.locator('id=tag1')).to_have_value('Hello world 3!')



from js import document
from pynanto.remote.components.ForwardTo import forward_to


class PnButton:

    def __init__(self, label: str, onclick=None):
        div = document.createElement("div")
        div.innerHTML = f"<button>{label}</button>"

        self._proxy = div.firstChild
        if onclick is not None:
            self.onclick = onclick
        document.body.append(div)


forward_to(PnButton, '_proxy')



from js import document
from pynanto.remote.components.ForwardTo import forward_to
from pynanto.remote.widget import Widget


class PnSlider:

    def __init__(self):
        w = Widget(f'<input type="range">')
        self._proxy = w.container.firstChild
        document.body.append(w.container)


forward_to(PnSlider, '_proxy')



import unittest

from js import document, HTMLElement
from pynanto.remote.unittest_fix import unittest_main_fixed
from pynanto.remote.widget import Widget, HolderWidget


class WidgetTestCase(unittest.TestCase):
    def test_basic_bind(self):
        class Widget1(Widget):
            def __init__(self):
                super().__init__("<span id='span1'>foo</span>")
                self.span1: HTMLElement = self
                self.bind_self_elements()

        target = Widget1()
        self.assertEqual('foo', target.span1.innerHTML)

    def test_bind_to_widget(self):
        class WidgetSub(Widget):
            def __init__(self):
                super().__init__("<div>I'm WidgetSub</div>")

        class Widget1(Widget):
            def __init__(self):
                super().__init__("<div id='div1'></div>")
                self.div1 = self(lambda: WidgetSub())

        target = Widget1()
        target.append_to(document.createElement('div'))
        html = target.container.innerHTML
        self.assertTrue("I'm WidgetSub" in html, f'Actual html=```{html}```')

    def test_bind_events(self):
        actual = []

        class Widget1(Widget):
            def __init__(self):
                super().__init__("<button id='foo'>foo</button>")

            def foo__click(self, *args):
                actual.append(1)

        target = Widget1()
        foo = target.container.querySelector('#foo')
        foo.click()
        self.assertEqual([1], actual)

    def test_w_type(self):
        class Widget1(Widget):
            def __init__(self):
                super().__init__('')

        target = Widget1()
        w_type = target.container.getAttribute('w-type')
        self.assertEqual('Widget1', w_type, f'outerHTML=`{target.container.outerHTML}`')

    def test_uninitialized(self):
        # GIVEN
        class Target(Widget):
            def __init__(self):
                super().__init__("<div id='second'></div>")
                self.first = self(lambda: Widget('orphan1'))
                self.second = self(lambda: Widget('non-orphan2'))

        target = Target()

        # WHEN
        call1 = target.uninitialized()
        actual1 = list(map(lambda wp: wp.name, call1))

        # THEN
        self.assertEqual(['first', 'second'], actual1)

        # WHEN
        ignore = target.container

        # THEN
        call2 = target.uninitialized()
        actual2 = list(map(lambda wp: wp.name, call2))
        self.assertEqual(['first'], actual2)

        # WHEN
        target.uninitialized_append_to(target.container)

        # THEN
        parts = target.container.innerHTML.split('second', 1)
        self.assertIn('orphan1', parts[1])


class WidgetHolderTestCase(unittest.TestCase):

    def test_holder(self):
        target = HolderWidget()
        target.show(Widget('one'))

        self.assertIn('one', target.container.innerHTML)

    def test_holder_twoWidgets(self):
        target = HolderWidget()
        target.show(Widget('one'))
        target.show(Widget('two'))

        self.assertNotIn('one', target.container.innerHTML)
        self.assertIn('two', target.container.innerHTML)

    def test_close(self):
        target = HolderWidget()
        target.show(Widget('one'))
        widget = Widget('two')
        target.show(widget)
        widget.close()
        self.assertIn('one', target.container.innerHTML)
        self.assertNotIn('two', target.container.innerHTML)


def main():
    unittest_main_fixed(__name__)
    # unittest.main()


if __name__ == '__main__':
    main()



# credits https://stackoverflow.com/questions/373335/how-do-i-get-a-cron-like-scheduler-in-python

import time
from datetime import datetime, timedelta
from typing import List

from pynanto.common.log_setup import getLoggerFor

logger = getLoggerFor(__file__)


class AllMatch(set):
    def __contains__(self, item): return True  # match everything

    def __str__(self) -> str: return '*'


_allMatch = AllMatch()


def _to_set(obj):
    if obj is None:
        return _allMatch
    if isinstance(obj, int):
        return {obj}  # Single item
    if not isinstance(obj, set):
        obj = set(obj)
    return obj


class Event(object):
    def __init__(self, action, sec=None, min=None, hour=None,
                 day=None, month=None, dow=None,
                 args=(), kwargs={}, log_action_run: bool = True):
        self.log_action_run = log_action_run
        self.secs = _to_set(sec)
        self.mins = _to_set(min)
        self.hours = _to_set(hour)
        self.days = _to_set(day)
        self.months = _to_set(month)
        self.dow = _to_set(dow)
        self.action = action
        self.args = args
        self.kwargs = kwargs

    def matchtime(self, t: datetime):
        """Return True if this event should trigger at the specified datetime"""
        return ((t.second in self.secs) and
                (t.minute in self.mins) and
                (t.hour in self.hours) and
                (t.day in self.days) and
                (t.month in self.months) and
                (t.weekday() in self.dow))

    def check(self, t: datetime):
        if not self.matchtime(t):
            return False
        if self.log_action_run:
            logger.info(f'Running job {self}')
        try:
            self.action(*self.args, **self.kwargs)
        except:
            logger.error(f'Error executing event {self}')
        return True

    def __str__(self) -> str:
        string = ' '.join(map(str, [self.secs, self.mins, self.hours, self.days, self.months, self.dow]))
        return string


class CronTab(object):
    def __init__(self, *events, log_loop: bool = False, log_executions: bool = False):
        self.log_loop = log_loop
        self.log_executions = log_executions
        self.events: List[Event] = events

    def run_thread(self, name='Cron'):
        import threading
        th = threading.Thread(target=self.run, daemon=True, name=name)
        th.start()

    def run(self):
        t = datetime(*datetime.now().timetuple()[:6])
        while True:

            t += timedelta(seconds=1)
            while datetime.now() < t:
                seconds = (t - datetime.now()).seconds
                if self.log_loop:
                    logger.info(f'sleeping for {seconds}')
                time.sleep(seconds)

            if self.log_loop:
                logger.info(f'checking events count {len(self.events)}')

            executed = []
            for e in self.events:
                if e.check(t):
                    executed.append(e)

            if self.log_executions and len(executed) > 0:
                logger.info(f'executed count {len(executed)}')
                for e in executed:
                    logger.info(f'  {e}')


def main():
    action = lambda arg: print(f' {datetime.now()} happened {arg}')
    event1 = Event(action, args=(1,))
    event2 = Event(action, sec={5, 10, 30, 45, 50}, args=(2,))
    CronTab(event1, event2).run()


if __name__ == '__main__':
    main()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path


@dataclass(frozen=True)
class Resource(ABC):
    arcname: str

    @abstractmethod
    def get_content(self) -> str:
        pass


@dataclass(frozen=True)
class StringResource(Resource):
    content: str

    def get_content(self) -> str:
        return self.content


@dataclass(frozen=True)
class PathResource(Resource):
    filepath: Path

    def get_content(self) -> str:
        return self.filepath.read_text()



from abc import ABC, abstractmethod
from time import sleep
from typing import Tuple

from pynanto.routes import Routes, Route
from pynanto.server import wait_url


class Webserver(ABC):
    def __init__(self):
        self.host: str = '0.0.0.0'
        self.port: int = 7777
        self.routes = Routes()

    def set_host(self, host: str) -> 'Webserver':
        self.host = host
        return self

    def set_port(self, port: int) -> 'Webserver':
        self.port = port
        return self

    def set_binding(self, binding: Tuple[str, int]) -> 'Webserver':
        self.set_host(binding[0])
        self.set_port(binding[1])
        return self

    def set_routes(self, routes: Routes) -> 'Webserver':
        if routes is None:
            raise Exception(f'Parameter routes cannot be None')
        self.routes = routes

        for route in self.routes.list:
            self._setup_route(route)

        return self

    def start_listen(self) -> 'Webserver':
        self._start_listen()
        return self

    @abstractmethod
    def _setup_route(self, route: Route):
        pass

    @abstractmethod
    def _start_listen(self):
        pass

    def wait_ready(self) -> 'Webserver':
        wait_url(self.localhost_url() + '/check_if_webserver_is_accepting_requests')
        return self

    def localhost_url(self) -> str:
        return f'http://127.0.0.1:{self.port}'


def wait_forever():
    while True:
        sleep(10)



from __future__ import annotations


def hash_args(content: str | None = None) -> dict:
    result = dict()
    if content is None:
        from js import window
        content = window.location.hash.removeprefix('#')
    for entry in content.split('&'):
        parts = entry.split('=', 1)
        key = parts[0]
        value = parts[1] if len(parts) > 1 else ''
        result[key] = value
    return result


def test_hash_args():
    target = hash_args('widget=W1')
    assert target == {'widget': 'W1'}


def test_hash_args_withNoValue():
    assert hash_args('noval') == {'noval': ''}
    assert hash_args('noval=') == {'noval': ''}


def test_hash_args_multiple():
    assert hash_args('a=1&b=2') == {'a': '1', 'b': '2'}



import inspect
from pathlib import Path
from typing import List, Optional

from pynanto.bundle import Bundle
from pynanto.bundle import strategy_no_cache, ResourceGenerator, LambdaBundle
from pynanto.common.bundle import bundle_definition
from pynanto.resource import StringResource
from pynanto.response import Response, Request


class Bundles:
    def __init__(self):
        self.list: List[Bundle] = []
        self.strategy = strategy_no_cache

    def to_response(self, request: Request) -> Response:
        return Response.application_zip(self.strategy(self.list))

    def add_resources(self, resource_generator: ResourceGenerator):
        self.list.append(LambdaBundle(resource_generator))

    def add_flat_folder(self, folder: Path, relative_to: Optional[Path] = None):
        self.add_resources(lambda: bundle_definition(folder, relative_to=relative_to))

    def add_file_content(self, filename: str, content: str):
        resource_list = [StringResource(filename, content)]
        self.list.append(LambdaBundle(lambda: resource_list))


def external_filename(stack_backtrack=1) -> Optional[Path]:
    pynanto_root = Path(__file__).resolve().parent

    for stack in inspect.stack():
        source_path = Path(stack.filename).resolve()
        if not path_is_contained(source_path, pynanto_root):
            stack_backtrack -= 1
            if stack_backtrack == 0:
                return source_path

    return None


def path_is_contained(child: Path, parent: Path):
    cl = len(child.parts)
    cp = len(parent.parts)
    if cl < cp:
        return False
    m = min(cl, cp)
    child_parts = child.parts[:m]
    parent_parts = parent.parts[:m]
    return child_parts == parent_parts



from __future__ import annotations

from typing import Dict


def groupby_to_dict(iterable, key) -> Dict:
    result = dict()
    for item in iterable:
        k = key(item)
        array = result.get(k, None)
        if array is None:
            array = []
            result[k] = array
        array.append(item)
    return result


def partition(iterable, key):
    sx = []
    dx = []
    for item in iterable:
        if key(item):
            sx.append(item)
        else:
            dx.append(item)
    return sx, dx


def remove_inplace(array, key):
    for idx in reversed(range(len(array))):
        st = array[idx]
        if key(st):
            del array[idx]



from __future__ import annotations

from typing import Callable, Awaitable

from js import window, console

import asyncio
from pyodide.ffi import create_once_callable


async def waitAnimationFrame():
    event = asyncio.Event()

    def callback(*args):
        console.log('waitAnimationFrame callback!')
        console.log(args)
        event.set()

    window.requestAnimationFrame(create_once_callable(callback))
    await event.wait()
    console.log('waitAnimationFrame done')


def set_timeout(callback: Callable[[], Awaitable[None]], timeout: int | None = 0):
    window.setTimeout(create_once_callable(callback), timeout)
